'''
Function:
    定义一些和数据相关的工具函数
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np


'''数据格式转换工具'''
class DataConverter():
    '''转元组类型'''
    @staticmethod
    def totuple(inp):
        if isinstance(inp, tuple) or isinstance(inp, list):
            return tuple(inp)
        if isinstance(inp, float): 
            inp = int(inp)
        if isinstance(inp, int): 
            inp = (inp, inp)
        else: 
            raise TypeError(f'data type {type(inp)} can not be converted to tuple')
        return inp
    '''转四元数'''
    @staticmethod
    def toquaternion(inp):
        if isinstance(inp, tuple) or isinstance(inp, list):
            assert len(inp) == 4
            return inp
        if isinstance(inp, float): 
            inp = int(inp)
        if isinstance(inp, int): 
            inp = (inp, inp, inp, inp)
        else:
            raise TypeError(f'data type {type(inp)} can not be converted to quaternion')
        return inp


'''initialize'''
from .linear import Linear
from .module import Module
from .flatten import Flatten
from .convolution import Conv2d
from .sequential import Sequential
from .criterion import MSELoss, CrossEntropy
from .activation import Softmax, Sigmoid, Tanh, ReLU, LeakyReLU, ELU, SELU, Softplus


'''
Function:
    定义激活函数
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np
from .module import Module


'''Softmax'''
class Softmax(Module):
    def __init__(self, dim=-1):
        super(Softmax, self).__init__(dim=dim)
    '''定义前向传播'''
    def forward(self, x):
        ex = np.exp(x - np.max(x, axis=self.dim, keepdims=True))
        return ex / np.sum(ex, axis=-1, keepdims=True)
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        p = self.forward(self.storage['x'])
        gradient = p * (1 - p)
        return accumulated_gradient * gradient


'''Sigmoid'''
class Sigmoid(Module):
    def __init__(self):
        super(Sigmoid, self).__init__()
    '''定义前向传播'''
    def forward(self, x):
        x = 1 / (1 + np.exp(-x))
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        p = self.forward(self.storage['x'])
        gradient = p * (1 - p)
        return accumulated_gradient * gradient


'''Tanh'''
class Tanh(Module):
    def __init__(self):
        super(Tanh, self).__init__()
    '''定义前向传播'''
    def forward(self, x):
        x = 2 / (1 + np.exp(-2 * x)) - 1
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        p = self.forward(self.storage['x'])
        gradient = 1 - np.power(p, 2)
        return accumulated_gradient * gradient


'''ReLU'''
class ReLU(Module):
    def __init__(self):
        super(ReLU, self).__init__()
    '''定义前向传播'''
    def forward(self, x):
        x = np.where(x >= 0, x, 0)
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        gradient = np.where(self.storage['x'] >= 0, 1, 0)
        return accumulated_gradient * gradient


'''LeakyReLU'''
class LeakyReLU(Module):
    def __init__(self, alpha=0.2):
        super(LeakyReLU, self).__init__(alpha=alpha)
    '''定义前向传播'''
    def forward(self, x):
        x = np.where(x >= 0, x, self.alpha * x)
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        gradient = np.where(self.storage['x'] >= 0, 1, self.alpha)
        return accumulated_gradient * gradient


'''ELU'''
class ELU(Module):
    def __init__(self, alpha=0.1):
        super(ELU, self).__init__(alpha=alpha)
    '''定义前向传播'''
    def forward(self, x):
        x = np.where(x >= 0.0, x, self.alpha * (np.exp(x) - 1))
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        gradient = np.where(self.storage['x'] >= 0.0, 1, self.alpha * np.exp(self.storage['x']))
        return accumulated_gradient * gradient


'''SELU'''
class SELU(Module):
    def __init__(self):
        super(SELU, self).__init__(
            alpha=1.6732632423543772848170429916717, scale=1.0507009873554804934193349852946
        )
    '''定义前向传播'''
    def forward(self, x):
        x = self.scale * np.where(x >= 0.0, x, self.alpha * (np.exp(x) - 1))
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        gradient = self.scale * np.where(self.storage['x'] >= 0.0, 1, self.alpha * np.exp(self.storage['x']))
        return accumulated_gradient * gradient


'''Softplus'''
class Softplus(Module):
    def __init__(self):
        super(Softplus, self).__init__()
    '''定义前向传播'''
    def forward(self, x):
        x = np.log(1 + np.exp(x))
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        gradient = 1 / (1 + np.exp(-x))
        return accumulated_gradient * gradient


'''
Function:
    实现Flatten层
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np
from .module import Module


'''Flatten层'''
class Flatten(Module):
    def __init__(self):
        super(Flatten, self).__init__()
    '''定义前向传播'''
    def forward(self, x):
        x = x.reshape((x.shape[0], -1))
        return x
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        accumulated_gradient = accumulated_gradient.reshape(self.storage['x'].shape)
        return accumulated_gradient


'''
Function:
    定义一些和图像相关的工具函数
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import numpy as np


'''图像格式转换工具'''
class ImageConverter():
    '''图像转column'''
    @staticmethod
    def im2col(image, kernel_size, stride, padding):
        padding = ((padding[0], padding[1]), (padding[2], padding[3]))
        image_padded = np.pad(image, ((0, 0), (0, 0), padding[0], padding[1]), mode='constant')
        k, i, j = ImageConverter.im2colindices(image.shape, kernel_size, stride, padding)
        cols = image_padded[:, k, i, j]
        cols = cols.transpose(1, 2, 0).reshape(kernel_size[0] * kernel_size[1] * image.shape[1], -1)
        return cols
    '''column转图像'''
    @staticmethod
    def col2im(cols, image_size, kernel_size, stride, padding):
        padding = ((padding[0], padding[1]), (padding[2], padding[3]))
        batch_size, num_channels, h, w = image_size
        image_padded = np.zeros((batch_size, num_channels, h + np.sum(padding[0]), w + np.sum(padding[1])))
        k, i, j = ImageConverter.im2colindices(image_size, kernel_size, stride, padding)
        cols = cols.reshape(num_channels * np.prod(kernel_size), -1, batch_size)
        cols = cols.transpose(2, 0, 1)
        np.add.at(image_padded, (slice(None), k, i, j), cols)
        pad_h, pad_w = padding
        return image_padded[:, :, pad_h[0]: h+pad_h[0], pad_w[0]: w+pad_w[0]]
    '''获得im2col的索引, Reference: CS231n Stanford'''
    @staticmethod
    def im2colindices(image_size, kernel_size, stride, padding):
        batch_size, num_channels, h, w = image_size
        out_h = int((h + np.sum(padding[0]) - kernel_size[0]) / stride[0] + 1)
        out_w = int((w + np.sum(padding[1]) - kernel_size[1]) / stride[1] + 1)
        # i0
        i0 = np.repeat(np.arange(kernel_size[0]), kernel_size[1])
        i0 = np.tile(i0, num_channels)
        # i1
        i1 = stride[0] * np.repeat(np.arange(out_h), out_w)
        # j0
        j0 = np.tile(np.arange(kernel_size[1]), kernel_size[0] * num_channels)
        # j1
        j1 = stride[1] * np.tile(np.arange(out_w), out_h)
        # i, j, k
        i = i0.reshape(-1, 1) + i1.reshape(1, -1)
        j = j0.reshape(-1, 1) + j1.reshape(1, -1)
        k = np.repeat(np.arange(num_channels), kernel_size[0] * kernel_size[1]).reshape(-1, 1)
        return (k, i, j)


'''
Function:
    实现卷积层
Author:
    Zhenchao Jin
微信公众号:
    Charles的皮卡丘
'''
import math
import numpy as np
from .module import Module
from ..utils import DataConverter, ImageConverter


'''2D卷积'''
class Conv2d(Module):
    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=True):
        super(Conv2d, self).__init__(
            in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
            stride=stride, padding=padding, bias=bias
        )
        # 格式为: (kernel_h, kernel_w)
        self.kernel_size = DataConverter.totuple(self.kernel_size)
        # 格式为: (stride_h, stride_w)
        self.stride = DataConverter.totuple(self.stride)
        # 格式为: (pad_h_top, pad_h_bottom, pad_w_left, pad_w_right)
        self.padding = DataConverter.toquaternion(self.padding)
        # 初始化权重
        thresh = 1 / math.sqrt(np.prod(self.kernel_size))
        self.weight = np.random.uniform(
            -thresh, thresh, (self.out_channels, self.in_channels, *self.kernel_size)
        )
        self.bias = np.zeros((self.out_channels, 1))
        # 初始化storage
        self.storage.update({
            'direction': {
                'weight': np.zeros(np.shape(self.weight)), 
                'bias': np.zeros(np.shape(self.bias))
            }
        })
    '''定义前向传播'''
    def forward(self, x):
        batch_size, num_channels, h, w = x.shape
        x_cols = ImageConverter.im2col(x, self.kernel_size, self.stride, self.padding)
        weight_cols = self.weight.reshape((self.out_channels, -1))
        self.storage.update({'x_cols': x_cols, 'weight_cols': weight_cols})
        feats = weight_cols.dot(x_cols) + self.bias
        feats = feats.reshape(self.outputsize(x)[1:] + (batch_size,)).transpose(3, 0, 1, 2)
        return feats
    '''定义反向传播'''
    def backward(self, accumulated_gradient):
        accumulated_gradient = accumulated_gradient.transpose(1, 2, 3, 0).reshape(self.out_channels, -1)
        if self.training:
            # 计算梯度
            grad_w = accumulated_gradient.dot(self.storage['x_cols'].T).reshape(self.weight.shape)
            grad_b = np.sum(accumulated_gradient, axis=1, keepdims=True)
            # 根据梯度更新weight
            results = self.update(self.weight, grad_w, self.storage['direction']['weight'])
            self.weight, self.storage['direction']['weight'] = results['params'], results['direction']
            # 根据梯度更新bias
            results = self.update(self.bias, grad_b, self.storage['direction']['bias'])
            self.bias, self.storage['direction']['bias'] = results['params'], results['direction']
        accumulated_gradient = self.storage['weight_cols'].T.dot(accumulated_gradient)
        accumulated_gradient = ImageConverter.col2im(accumulated_gradient, self.storage['x'].shape, self.kernel_size, self.stride, self.padding)
        return accumulated_gradient
    '''返回参数数量'''
    def parameters(self):
        return np.prod(self.weight.shape) + np.prod(self.bias.shape)
    '''特征输出大小'''
    def outputsize(self, inp):
        batch_size, num_channels, h, w = inp.shape
        output_size = (
            batch_size,
            self.out_channels,
            int((h + np.sum(self.padding[:2]) - self.kernel_size[0]) / self.stride[0] + 1),
            int((w + np.sum(self.padding[2:]) - self.kernel_size[1]) / self.stride[1] + 1),
        )
        return output_size